#!/usr/bin/perl
#
# Copyright oota
#
# License GPL2 or Later
#
# Ver 0.15
use warnings;
use strict;

use IO::Socket;
use Data::Hexdumper;
use File::Log;
use Net::Interface;

sub put_name;
sub put_nmb_name;
sub convert_str_to_ip;

# Data definition
#
# Constant
#
      eval 'sub MAX_NETBIOSNAME_LEN ()  {16;} ';
      eval 'sub WINSPORT ()		{137;}';
      my $debuglevel		= 5;
# please edit this line
      my $config_file	= "/etc/samba/winslist.conf";
      my $logfile	= "/var/log/samba/winssend.log";
#
# Value'
#
#   loop value
#
       my $i;
#   parameter
      my $param_mode;
      my $param_name;
      my $param_type;
      my $param_ttl;
      my $param_ip;
#
#   config data
#
      my $config_dir;
      my @wins_server_list;
#
#   File::Log handle value and debug value
#
      my $loghandle;
      my $dumpdata;
#
#   Net::Interface handle value and host/netmask value
#
      my @All_ifs;
      my $intname;             #interface name
      my $if;		       #intefrace handle
      my $naddr;	       #address data
      my $hostaddr;            #my host address
      my $netmask;             #my netmask
      my $hostmask;            #masked host address
      my $clientaddr;	       #client address
      my $clientmask;          #masked client address
      my $sameflag;	       #same segment or not
#
#  nmb packet from RFC1002
#
      my $sock;			#socket
      my $nmb_packet;		#send data
      my $nmb_header;		#header data
      my $additional_rec;	#additional data

      my $name_trn_id;
      my $opcode;
      my $nm_flag;
      my $rcode;
      my $qd_count;
      my $an_count;
      my $ns_count;
      my $ar_count;
      my $query_name;
      my $query_type;
      my $query_class;
      my $additional_rec_header;
      my $additional_rec_name;
      my $additional_rec_type;
      my $additional_rec_class;
      my $additional_rec_ttl;
      my $additional_rec_length;
      my $additional_rec_flag;
      my $additional_rec_addr;
#
# nmb packet structure
#
#
# header-part
# NAME_TRN_ID		#network byte order short
# OPCODE		#opcode(5bit)
			# top bit (0:request/1:response)
			# Registration (0x05)
			# Release (0x06)
			# Refresh (0x08)
			# Multi-Homed Name Registration (0x0f)
# NM_FLAG		#7bit
#   AA			#  Authoritative Answer flag
#   TC			#  Truncation Flag
#   RD			#  Recursion Desired Flag
#   RA			#  Recursion Available Flag
#   0			#  always 0
#   0			#  always 0
#   B			#  Broadcast Flag
#
# RCODE			#4bit Result codes of request
#
# QD_COUNT		#network byte order short
#			#  Question section entries
# AN_COUNT		#network byte order short
#			#  Answer section entries
# NS_COUNT		#network byte order short
#			#  Authority section entries
# AR_COUNT		#network byte order short
#			#  Additional section entries
#

#
# open log
#
  $loghandle = File::Log->new(  debug=>$debuglevel,
				logFileName=>$logfile,
				logFileDateTime=>0,
				dateTimeStamp=>1,
				appName=>'winssend');

#
# get wins server list
#
  open WI,$config_file or die "winslist.conf not found.\n";
  while(<WI>) {
    if($_ =~ /wins list = /) {
        split;
	splice(@_,0,3);
        @wins_server_list = @_;
    }
  }
  close WI;
  $loghandle->msg(1,"start winssend\n");
#
# get parameter
#
  $param_mode = $ARGV[0];
  $param_name = $ARGV[1];
  $param_type = $ARGV[2];
  $param_ttl  = $ARGV[3];
  $param_ip   = $ARGV[4];
  my $argstr = "add BPNT3 03 259200 192.168.1.15";
  ($param_mode,$param_name,$param_type,$param_ttl,$param_ip) = split / /,$argstr;
  $loghandle->msg(1,"register data=$param_mode $param_name $param_type" . 
                    " $param_ttl $param_ip\n");
  $clientaddr = convert_str_to_number($param_ip);
#
# check client address is same segment or not
#
  @All_ifs = Net::Interface->interfaces();
  $sameflag = 0;
  foreach(@All_ifs) {
      $intname = $_->name;
      next if $intname eq "lo";
      print $_->name,"\n";
      $if         = Net::Interface->new ($intname);
      $naddr      = $if->address();
      $hostaddr   = convert_ip_to_number($naddr);
      $naddr      = $if->netmask();
      $netmask	  = convert_ip_to_number($naddr);
      $hostmask   = $hostaddr   & $netmask;
      $clientmask = $clientaddr & $netmask;
#      printf "\nmask=%8x:%8x\n",$hostmask,$clientmask;
      if( $hostmask != $clientmask ) {
         $dumpdata = sprintf "host/crient/mask=%8x:%8x:%8x not same!\n",
			$hostaddr,$clientaddr,$netmask;
         $loghandle->msg(1,$dumpdata);
         $sameflag = 1;
      }
   };
   exit if $sameflag == 1;
#
# construct NBT packet
#
  $name_trn_id = $$;		#process no
  if($param_mode eq "add") {
    $opcode	= 0x0f; 	#multihomed registration
  }elsif($param_mode eq "refresh") {
    $opcode	= 0x08;		#refresh
  }elsif($param_mode eq "del") {
    $opcode	= 0x06;		#release
  }
  $nm_flag	= 0x10;		#not truncated,
				#Do query recursively
				#not a broadcast packet
  $rcode	= 0;
  $qd_count	= 1;		#query exist
  $an_count	= 0;
  $ns_count	= 0;
  $ar_count	= 1;		#additional record exist
  $query_name	= $param_name;
  $query_type	= 0x20;
  $query_class	= 1;
  $additional_rec_header = 0xc0;
  $additional_rec_name   = 0x0c;
  $additional_rec_type   = 0x20;
  $additional_rec_class  = 1;
  $additional_rec_ttl    = $param_ttl;
  $additional_rec_length = 6;
  $additional_rec_flag   = 0;
  $additional_rec_addr   = convert_str_to_ip($param_ip);
#
# send
#
#   pack packet
#        header
  $nmb_header = pack "n" , $name_trn_id;
  $nmb_header = $nmb_header . pack "n" , ($opcode <<11) + ($nm_flag << 4) + $rcode;
  $nmb_header = $nmb_header . pack "nnnn" , $qd_count,$an_count,$ns_count,$ar_count;
#
#  print hexdump(data => $nmb_header,number => 'C' );
#	name part
  $query_name = put_nmb_name($param_name,$param_type);
  $nmb_header = $nmb_header . $query_name;
#  print hexdump(data => $nmb_header,number => 'C' );
#	additinoal parts
  $additional_rec = pack "nn",$query_type,$query_class;
  $additional_rec = $additional_rec . pack "CC",$additional_rec_header,$additional_rec_name;
  $additional_rec = $additional_rec . pack "nn",$additional_rec_type,$additional_rec_class;
  $additional_rec = $additional_rec . pack "N" ,$additional_rec_ttl;
  $additional_rec = $additional_rec . pack "nn",$additional_rec_length,$additional_rec_flag;
  $additional_rec = $additional_rec . $additional_rec_addr;
#	all 
  $nmb_packet = $nmb_header . $additional_rec;
  $dumpdata = hexdump(data => $nmb_packet,number=>'C');
  $loghandle->msg(4,"socket data =\n",
		   $dumpdata,
		   "\nwins server list=",@wins_server_list,"\n");
#  exit;
#
#   socket make
#
  foreach $i (@wins_server_list) {
    $sock=();
    $sock = new IO::Socket::INET(PeerAddr => $i,
                             PeerPort => &WINSPORT,
                             Proto    => 'udp')
        or die "Can not create socket. Reason: $!\n" unless $sock;
#
#   send udp data and close socket
#
    $loghandle->msg(4,"PeerAddr= $i\n");
    print $sock $nmb_packet;
    $sock->flush();
    close($sock);
  }
  $loghandle->msg(1,"winssend end\n");
#
# translate nmb name to packet format
# See RFC1001 and samba source/libsmb/nmblib.c
#
sub put_name {
  my($name,$pad,$name_type) = @_;
  my $len;
  my $outbuf;
  my $i;
  my @arraybuf1;
  my $outchar;

  @arraybuf1 = split //,$name_type;
  $len = length($name);
  if($len < &MAX_NETBIOSNAME_LEN -1 ) {
    $outbuf = $name . $pad x (&MAX_NETBIOSNAME_LEN - 1 - $len);
  }
  $outchar = (ord(@arraybuf1[0])-48)*16;
  $outchar = $outchar+(ord(@arraybuf1[1])-48);
  $outbuf = $outbuf . chr($outchar);
  return $outbuf;
}

sub put_nmb_name{
  my($name,$name_type) = @_;
  my $buf1;
  my @buf;
  my @arraybuf1;
  my $ptr;

  if($name eq "*") {
    $buf1 = put_name("*","\0",$name_type);
  } else {
    $buf1 = put_name($name," ",$name_type);
  }
#convert string to array
   @arraybuf1 = split //,$buf1;
#convert string to digit
  $buf[0] = chr(0x20);
  for($i=0;$i<&MAX_NETBIOSNAME_LEN;$i = $i+1){
    @buf[1+2*$i] = chr(ord('A') + ((ord(@arraybuf1[$i]) >>4) & 0xf));
    @buf[2+2*$i] = chr(ord('A') + (ord(@arraybuf1[$i]) & 0xf));
  }
  $buf[33] = chr(0x0);
  return join "",@buf;
}

#
#convert string to 4digit ip address
#
sub convert_str_to_ip {
  my @ip;
  @ip = split /\./,$_[0];
  return pack "CCCC",$ip[0],$ip[1],$ip[2],$ip[3]
}
#
sub convert_str_to_number {
  my @ip;
  @ip = split /\./,$_[0];
  return (($ip[0]*256+$ip[1])*256+$ip[2])*256+$ip[3]
}
#
sub convert_ip_to_number {
  my @ip;
  @ip = unpack( "CCCC",$_[0]);
  return (($ip[0]*256+$ip[1])*256+$ip[2])*256+$ip[3];
}
#
#ChangeLog
#
#Ver 0.01   2006/02/14 Initial Release.
#Ver 0.02   2006/02/22 Change testparm -vs to direct.
#	               Use File::Log.
#Ver 0.11   2006/05/22 Check Same network host or not.
#                      multiple wins server support.
#                      Use Net::Interface.
#Ver 0.12   2006/05/24 Add starting message.
#
#Ver 0.13   2007/04/27 Name type wrong set bug fix.
#
#Ver 0.14   2007/07/12 winslist.conf file analysis bug fix.
#
#Ver 0.15   2007/07/16 winslist.conf parameter analysis bug fix.
#                      socket initialize bug fix.
                       Change Licence (GPL2 only -> GPL2 or later)
